# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _layer_normalization(X, W, B, axis=-1, epsilon=1e-5):  # type: ignore
    # Inspired from: https://github.com/onnx/onnx/blob/main/onnx/backend/
    # test/case/node/layernormalization.py#L12
    X_shape = X.shape
    X_rank = len(X_shape)
    if axis < 0:
        # If axis = -1 and rank of X is 4,
        # the axis is changed to -1 + 4 = 3,
        # which means the last axis.
        axis = axis + X_rank
    unsqueezed_rank = X_rank - axis
    reduction_shape = X_shape[0:axis] + (1,) * unsqueezed_rank

    # Parameter used to convert N-D tensor layer
    # normalization to equivalent 2-D matirx operations.
    row_number = 1
    col_number = 1
    for i in range(X_rank):
        if i < axis:
            row_number *= X_shape[i]
        else:
            col_number *= X_shape[i]

    # After reshaping input tensor X into a matrix,
    # layer normalization is equivalent to conducting
    # standardization on each column vector (s.t. each
    # column has zero mean and unit variance).
    x_mat = numpy.reshape(X, (row_number, col_number))
    # This computes mean for every x_mat's column.
    x_mean = numpy.sum(x_mat, axis=1, keepdims=True) / col_number
    x_diff = x_mat - x_mean
    x_squared_diff = x_diff * x_diff
    # This computes variance for every x_mat's column.
    variance = numpy.sum(x_squared_diff, axis=1, keepdims=True) / col_number
    variance_eps = variance + epsilon
    std_dev = numpy.sqrt(variance_eps)
    inv_std_dev = numpy.reciprocal(std_dev)
    # Standardization step. y_mat is zero-mean and unit-variance.
    y_mat = x_diff * inv_std_dev
    # Apply affine transform on normalization outcome.
    # W is linear coefficient while B is bias.
    Y = numpy.reshape(y_mat, X_shape) * W
    if B is not None:
        Y = Y + B
    # Matrix-level operations' outputs should be reshaped
    # to compensate the initial tensor-to-matrix reshape.
    X_mean = numpy.reshape(x_mean, reduction_shape)
    X_inv_std_dev = numpy.reshape(inv_std_dev, reduction_shape)

    return (Y.astype(X.dtype),
            X_mean.astype(X.dtype),
            X_inv_std_dev.astype(X.dtype))


class LayerNormalization(OpRun):

    atts = {'axis': -1,
            'epsilon': 9.999999747378752e-06,
            'stash_type': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=LayerNormalization.atts,
                       **options)

    def _run(self, X, Scale, B=None, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        res = _layer_normalization(
            X, Scale, B, axis=self.axis, epsilon=self.epsilon)
        return res



"""
@file
@brief Parsers for LightGBM booster.
"""
import numpy
from sklearn.base import ClassifierMixin
from skl2onnx._parse import _parse_sklearn_classifier, _parse_sklearn_simple_model
from skl2onnx.common._apply_operation import apply_concat, apply_cast
from skl2onnx.common.data_types import guess_proto_type
from skl2onnx.proto import onnx_proto


class WrappedLightGbmBooster:
    """
    A booster can be a classifier, a regressor.
    Trick to wrap it in a minimal function.
    """

    def __init__(self, booster):
        self.booster_ = booster
        self.n_features_ = self.booster_.feature_name()
        self.objective_ = self.get_objective()
        if self.objective_.startswith('binary'):
            self.operator_name = 'LgbmClassifier'
            self.classes_ = self._generate_classes(booster)
        elif self.objective_.startswith('multiclass'):
            self.operator_name = 'LgbmClassifier'
            self.classes_ = self._generate_classes(booster)
        elif self.objective_.startswith('regression'):  # pragma: no cover
            self.operator_name = 'LgbmRegressor'
        else:  # pragma: no cover
            raise NotImplementedError(
                f'Unsupported LightGbm objective: {self.objective_!r}.')
        try:
            bt = self.booster_.attr('boosting_type')
        except KeyError:
            bt = None
        if bt is None:
            try:
                bt = self.booster_.params['boosting_type']
            except AttributeError:
                bt = 'gbdt'
        self.boosting_type = bt
        # if average_output:
        #     self.boosting_type = 'rf'
        # else:
        #     self.boosting_type = 'gbdt'

    @staticmethod
    def _generate_classes(booster):
        if isinstance(booster, dict):
            num_class = booster['num_class']
        else:
            num_class = booster.attr('num_class')
        if num_class is None:
            dp = booster.dump_model(num_iteration=1)
            num_class = dp['num_class']
        if num_class == 1:
            return numpy.asarray([0, 1])
        return numpy.arange(num_class)

    def get_objective(self):
        "Returns the objective."
        if hasattr(self, 'objective_') and self.objective_ is not None:
            return self.objective_
        objective = self.booster_.attr('objective')
        if objective is not None:
            return objective
        dp = self.booster_.dump_model(num_iteration=1)
        return dp['objective']


class WrappedLightGbmBoosterClassifier(ClassifierMixin):
    """
    Trick to wrap a LGBMClassifier into a class.
    """

    def __init__(self, wrapped):  # pylint: disable=W0231
        for k in {'boosting_type', '_model_dict', '_model_dict_info',
                  'operator_name', 'classes_', 'booster_', 'n_features_',
                  'objective_'}:
            if hasattr(wrapped, k):
                setattr(self, k, getattr(wrapped, k))


class MockWrappedLightGbmBoosterClassifier(WrappedLightGbmBoosterClassifier):
    """
    Mocked lightgbm.
    """

    def __init__(self, tree):  # pylint: disable=W0231
        self.dumped_ = tree
        self.is_mock = True

    def dump_model(self):
        "mock dump_model method"
        self.visited = True
        return self.dumped_

    def feature_name(self):
        "Returns binary features names."
        return [0, 1]

    def attr(self, key):
        "Returns default values for common attributes."
        if key == 'objective':
            return "binary"
        if key == 'num_class':
            return 1
        if key == 'average_output':
            return None
        raise KeyError(  # pragma: no cover
            f"No response for {key!r}.")


def lightgbm_parser(scope, model, inputs, custom_parsers=None):
    """
    Agnostic parser for LightGBM Booster.
    """
    if hasattr(model, "fit"):
        raise TypeError(  # pragma: no cover
            f"This converter does not apply on type '{type(model)}'.")

    if len(inputs) == 1:
        wrapped = WrappedLightGbmBooster(model)
        objective = wrapped.get_objective()
        if objective.startswith('binary'):
            wrapped = WrappedLightGbmBoosterClassifier(wrapped)
            return _parse_sklearn_classifier(
                scope, wrapped, inputs, custom_parsers=custom_parsers)
        if objective.startswith('multiclass'):
            wrapped = WrappedLightGbmBoosterClassifier(wrapped)
            return _parse_sklearn_classifier(
                scope, wrapped, inputs, custom_parsers=custom_parsers)
        if objective.startswith('regression'):  # pragma: no cover
            return _parse_sklearn_simple_model(
                scope, wrapped, inputs, custom_parsers=custom_parsers)
        raise NotImplementedError(  # pragma: no cover
            f"Objective '{objective}' is not implemented yet.")

    # Multiple columns
    this_operator = scope.declare_local_operator('LightGBMConcat')
    this_operator.raw_operator = model
    this_operator.inputs = inputs
    var = scope.declare_local_variable(
        'Xlgbm', inputs[0].type.__class__([None, None]))
    this_operator.outputs.append(var)
    return lightgbm_parser(scope, model, this_operator.outputs,
                           custom_parsers=custom_parsers)


def shape_calculator_lightgbm_concat(operator):
    """
    Shape calculator for operator *LightGBMConcat*.
    """
    pass


def converter_lightgbm_concat(scope, operator, container):
    """
    Converter for operator *LightGBMConcat*.
    """
    op = operator.raw_operator
    options = container.get_options(op, dict(cast=False))
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:  # pylint: disable=E1101
        proto_dtype = onnx_proto.TensorProto.FLOAT  # pylint: disable=E1101
    if options['cast']:
        concat_name = scope.get_unique_variable_name('cast_lgbm')
        apply_cast(scope, concat_name, operator.outputs[0].full_name, container,
                   operator_name=scope.get_unique_operator_name('cast_lgmb'),
                   to=proto_dtype)
    else:
        concat_name = operator.outputs[0].full_name

    apply_concat(scope, [_.full_name for _ in operator.inputs],
                 concat_name, container, axis=1)



"""
@file
@brief Rewrites some of the converters implemented in
:epkg:`sklearn-onnx`.
"""
from sklearn.preprocessing import FunctionTransformer
from skl2onnx.algebra.onnx_operator import OnnxSubEstimator


def transformer_target_regressor_shape_calculator(operator):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support custom functions
    implemented with :ref:`l-numpy-onnxpy`.
    """
    input_type = operator.inputs[0].type.__class__
    # same output shape as input
    output_type = input_type([None, None])
    operator.outputs[0].type = output_type


def transformer_target_regressor_converter(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support custom functions
    implemented with :ref:`l-numpy-onnxpy`.
    """
    op = operator.raw_operator
    opv = container.target_opset
    X = operator.inputs[0]

    Y = OnnxSubEstimator(op.regressor_, X, op_version=opv)
    cpy = FunctionTransformer(op.transformer_.inverse_func)
    Z = OnnxSubEstimator(cpy, Y, output_names=operator.outputs)
    Z.add_to(scope, container)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from .op_dft import _cfft as _dft
from .op_slice import _slice
from .op_concat_from_sequence import _concat_from_sequence


def _concat(*args, axis=0):
    return numpy.concatenate(tuple(args), axis=axis)


def _unsqueeze(a, axis):
    return numpy.expand_dims(a, axis=axis)


def _switch_axes(a, ax1, ax2):
    p = [i for i in range(len(a.shape))]
    p[ax1], p[ax2] = p[ax2], p[ax1]
    return numpy.transpose(a, p)


def _stft(x, fft_length, hop_length, n_frames, window, onesided=False):
    """
    Applies one dimensional FFT with window weights.
    torch defines the number of frames as:
    `n_frames = 1 + (len - n_fft) / hop_length`.
    """
    last_axis = len(x.shape) - 1  # op.Sub(op.Shape(op.Shape(x)), one)
    axis = [-2]
    axis2 = [-3]
    window_size = window.shape[0]

    # building frames
    seq = []
    for fs in range(n_frames):
        begin = fs * hop_length
        end = begin + window_size
        sliced_x = _slice(x, numpy.array([begin]), numpy.array([end]), axis)

        # sliced_x may be smaller
        new_dim = sliced_x.shape[-2:-1]
        missing = (window_size - new_dim[0], )
        new_shape = sliced_x.shape[:-2] + missing + sliced_x.shape[-1:]
        cst = numpy.zeros(new_shape, dtype=x.dtype)
        pad_sliced_x = _concat(sliced_x, cst, axis=-2)

        # same size
        un_sliced_x = _unsqueeze(pad_sliced_x, axis2)
        seq.append(un_sliced_x)

    # concatenation
    new_x = _concat_from_sequence(seq, axis=-3, new_axis=0)

    # calling weighted dft with weights=window
    shape_x = new_x.shape
    shape_x_short = shape_x[:-2]
    shape_x_short_one = tuple(1 for _ in shape_x_short) + (1, )
    window_shape = shape_x_short_one + (window_size, 1)
    weights = numpy.reshape(window, window_shape)
    weighted_new_x = new_x * weights

    result = _dft(weighted_new_x, fft_length, last_axis,
                  onesided=onesided)  # normalize=False

    # final transpose -3, -2
    dim = len(result.shape)
    ax1 = dim - 3
    ax2 = dim - 2
    return _switch_axes(result, ax1, ax2)


def _istft(x, fft_length, hop_length, window, onesided=False):  # pylint: disable=R0914
    """
    Reverses of `stft`.
    """
    zero = [0]
    one = [1]
    two = [2]
    axisf = [-2]
    n_frames = x.shape[-2]
    expected_signal_len = fft_length[0] + hop_length * (n_frames - 1)

    # building frames
    seqr = []
    seqi = []
    seqc = []
    for fs in range(n_frames):
        begin = fs
        end = fs + 1
        frame_x = numpy.squeeze(_slice(x, numpy.array([begin]),
                                       numpy.array([end]), axisf),
                                axis=axisf[0])

        # ifft
        ift = _dft(frame_x, fft_length, axis=-1, onesided=onesided,
                   normalize=True)
        n_dims = len(ift.shape)

        # real part
        n_dims_1 = n_dims - 1
        sliced = _slice(ift, numpy.array(zero),
                        numpy.array(one), [n_dims_1])
        ytmp = numpy.squeeze(sliced, axis=n_dims_1)
        ctmp = numpy.full(ytmp.shape, fill_value=1, dtype=x.dtype) * window

        shape_begin = ytmp.shape[:-1]
        n_left = fs * hop_length
        size = ytmp.shape[-1]
        n_right = expected_signal_len - (n_left + size)

        left_shape = shape_begin + (n_left, )
        right_shape = shape_begin + (n_right, )
        right = numpy.zeros(right_shape, dtype=x.dtype)
        left = numpy.zeros(left_shape, dtype=x.dtype)

        y = _concat(left, ytmp, right, axis=-1)
        yc = _concat(left, ctmp, right, axis=-1)

        # imaginary part
        sliced = _slice(ift, numpy.array(one), numpy.array(two), [n_dims_1])
        itmp = numpy.squeeze(sliced, axis=n_dims_1)
        yi = _concat(left, itmp, right, axis=-1)

        # append
        seqr.append(_unsqueeze(y, axis=-1))
        seqi.append(_unsqueeze(yi, axis=-1))
        seqc.append(_unsqueeze(yc, axis=-1))

    # concatenation
    redr = _concat_from_sequence(seqr, axis=-1, new_axis=0)
    redi = _concat_from_sequence(seqi, axis=-1, new_axis=0)
    redc = _concat_from_sequence(seqc, axis=-1, new_axis=0)

    # unweight
    resr = redr.sum(axis=-1, keepdims=0)
    resi = redi.sum(axis=-1, keepdims=0)
    resc = redc.sum(axis=-1, keepdims=0)
    rr = resr / resc
    ri = resi / resc

    # Make complex
    rr0 = numpy.expand_dims(rr, axis=0)
    ri0 = numpy.expand_dims(ri, axis=0)
    conc = _concat(rr0, ri0, axis=0)

    # rotation, bring first dimension to the last position
    result_shape = conc.shape
    reshaped_result = conc.reshape((2, -1))
    transposed = numpy.transpose(reshaped_result, (1, 0))
    other_dimensions = result_shape[1:]
    final_shape = _concat(other_dimensions, two, axis=0)
    final = transposed.reshape(final_shape)
    return final


class STFT(OpRun):

    atts = {'onesided': 1, 'inverse': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=STFT.atts,
                       **options)

    def _run(self, x, frame_step, window=None, frame_length=None,  # pylint: disable=W0221
             attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if frame_length is None:
            frame_length = x.shape[-2]
        hop_length = frame_length // 4
        if window is None:
            window = numpy.ones(x.shape[-2], dtype=x.dtype)
        if self.inverse:
            res = _istft(x, [frame_length], hop_length, window,
                         onesided=self.onesided)
        else:
            n_frames = 1  # int(1 + (x.shape[-2] - frame_length) / hop_length)
            res = _stft(x, [frame_length], hop_length, n_frames, window,
                        onesided=self.onesided)
        return (res.astype(x.dtype), )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.onnx_pb import TensorProto
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE
from ._op import OpRun


class _CommonWindow:

    def _begin(self, size):
        if self.periodic == 1:
            N_1 = size
        else:
            N_1 = size - 1
        ni = numpy.arange(size, dtype=self.dtype)
        return ni, N_1

    def _end(self, size, res):
        return (res.astype(self.dtype), )


class BlackmanWindow(OpRun, _CommonWindow):
    """
    Returns
    :math:`\\omega_n = 0.42 - 0.5 \\cos \\left( \\frac{2\\pi n}{N-1} \\right) +
    0.08 \\cos \\left( \\frac{4\\pi n}{N-1} \\right)`
    where *N* is the window length.
    See `blackman_window
    <https://pytorch.org/docs/stable/generated/torch.blackman_window.html>`_
    """

    atts = {'output_datatype': TensorProto.FLOAT, 'periodic': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=BlackmanWindow.atts,
                       **options)
        self.dtype = TENSOR_TYPE_TO_NP_TYPE[self.output_datatype]

    def _run(self, size, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        # ni, N_1 = self._begin(size)
        ni, N_1 = numpy.arange(size, dtype=self.dtype), size
        alpha = 0.42
        beta = 0.08
        pi = 3.1415
        y = alpha
        y -= numpy.cos((ni * (pi * 2)) / N_1) / 2
        y += numpy.cos((ni * (pi * 4)) / N_1) * beta
        return (self._end(size, y), )


class HannWindow(OpRun, _CommonWindow):
    """
    Returns
    :math:`\\omega_n = \\sin^2\\left( \\frac{\\pi n}{N-1} \\right)`
    where *N* is the window length.
    See `hann_window
    <https://pytorch.org/docs/stable/generated/torch.hann_window.html>`_
    """

    atts = {'output_datatype': TensorProto.FLOAT, 'periodic': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=HannWindow.atts,
                       **options)
        self.dtype = TENSOR_TYPE_TO_NP_TYPE[self.output_datatype]

    def _run(self, size, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        ni, N_1 = self._begin(size)
        res = numpy.sin(ni * 3.1415 / N_1) ** 2
        return self._end(size, res)


class HammingWindow(OpRun, _CommonWindow):
    """
    Returns
    :math:`\\omega_n = \\alpha - \\beta \\cos \\left( \\frac{\\pi n}{N-1} \\right)`
    where *N* is the window length.
    See `hamming_window
    <https://pytorch.org/docs/stable/generated/torch.hamming_window.html>`_.
    `alpha=0.54, beta=0.46`
    """

    atts = {'output_datatype': TensorProto.FLOAT, 'periodic': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=HammingWindow.atts,
                       **options)
        self.dtype = TENSOR_TYPE_TO_NP_TYPE[self.output_datatype]

    def _run(self, size, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        ni, N_1 = self._begin(size)
        alpha = 25. / 46.
        beta = 1 - alpha
        res = alpha - numpy.cos(ni * 3.1415 * 2 / N_1) * beta
        return self._end(size, res)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class OptionalGetElement(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc, **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if not isinstance(x, list):
            raise TypeError(  # pragma: no cover
                f"Unexpected type {type(x)!r} for x.")
        if len(x) > 0:
            return (x[0], )
        return ([], )


class OptionalHasElement(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc, **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if not isinstance(x, list):
            raise TypeError(  # pragma: no cover
                f"Unexpected type {type(x)!r} for x.")
        if len(x) > 0:
            return (numpy.array([e is not None for e in x]), )
        return ([], )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _concat_from_sequence(seq, axis, new_axis=0):
    if new_axis == 1:
        seq2 = [s[..., numpy.newaxis] for s in seq]
        res = numpy.concatenate(seq2, axis=-1)
    else:
        res = numpy.concatenate(seq, axis=axis)
    return res


class ConcatFromSequence(OpRun):

    atts = {'axis': 0, 'new_axis': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=ConcatFromSequence.atts,
                       **options)

    def _run(self, seq, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if seq is None:
            raise RuntimeError(  # pragma: no cover
                "A sequence cannot be null.")
        res = _concat_from_sequence(seq, self.axis, new_axis=self.new_axis)
        return (res, )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _fft(x, fft_length, axis):
    if fft_length is None:
        fft_length = [x.shape[axis]]
    ft = numpy.fft.fft(x, fft_length[0], axis=axis)
    r = numpy.real(ft)
    i = numpy.imag(ft)
    merged = numpy.vstack([r[numpy.newaxis, ...], i[numpy.newaxis, ...]])
    perm = numpy.arange(len(merged.shape))
    perm[:-1] = perm[1:]
    perm[-1] = 0
    tr = numpy.transpose(merged, list(perm))
    if tr.shape[-1] != 2:
        raise RuntimeError(
            f"Unexpected shape {tr.shape}, x.shape={x.shape} "
            f"fft_length={fft_length}.")
    return tr


def _cfft(x, fft_length, axis, onesided=False, normalize=False):
    # if normalize:
    #    raise NotImplementedError()
    if x.shape[-1] == 1:
        tmp = x
    else:
        slices = [slice(0, x) for x in x.shape]
        slices[-1] = slice(0, x.shape[-1], 2)
        real = x[tuple(slices)]
        slices[-1] = slice(1, x.shape[-1], 2)
        imag = x[tuple(slices)]
        tmp = real + 1j * imag
    c = numpy.squeeze(tmp, -1)
    res = _fft(c, fft_length, axis=axis)
    if onesided:
        slices = [slice(0, a) for a in res.shape]
        slices[axis] = slice(0, res.shape[axis] // 2 + 1)
        return res[tuple(slices)]
    return res


def _ifft(x, fft_length, axis=-1, onesided=False):
    ft = numpy.fft.ifft(x, fft_length[0], axis=axis)
    r = numpy.real(ft)
    i = numpy.imag(ft)
    merged = numpy.vstack([r[numpy.newaxis, ...], i[numpy.newaxis, ...]])
    perm = numpy.arange(len(merged.shape))
    perm[:-1] = perm[1:]
    perm[-1] = 0
    tr = numpy.transpose(merged, list(perm))
    if tr.shape[-1] != 2:
        raise RuntimeError(
            f"Unexpected shape {tr.shape}, x.shape={x.shape} "
            f"fft_length={fft_length}.")
    if onesided:
        slices = [slice() for a in tr.shape]
        slices[axis] = slice(0, tr.shape[axis] // 2 + 1)
        return tr[tuple(slices)]
    return tr


def _cifft(x, fft_length, axis=-1, onesided=False):
    if x.shape[-1] == 1:
        tmp = x
    else:
        slices = [slice(0, x) for x in x.shape]
        slices[-1] = slice(0, x.shape[-1], 2)
        real = x[tuple(slices)]
        slices[-1] = slice(1, x.shape[-1], 2)
        imag = x[tuple(slices)]
        tmp = real + 1j * imag
    c = numpy.squeeze(tmp, -1)
    return _ifft(c, fft_length, axis=axis, onesided=onesided)


class DFT(OpRun):

    atts = {'axis': 1, 'inverse': 0, 'onesided': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=DFT.atts,
                       **options)

    def _run(self, x, dft_length=None, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if dft_length is None:
            dft_length = numpy.array([x.shape[self.axis]], dtype=numpy.int64)
        if self.inverse:
            res = _cifft(x, dft_length, axis=self.axis, onesided=self.onesided)
        else:
            res = _cfft(x, dft_length, axis=self.axis, onesided=self.onesided)
        return (res.astype(x.dtype), )
